var width = 960,
    height = 500;

var nodes = d3.range(1, 5).map(function(i) {
    return [i * width / 5, 50 + Math.random() * (height - 100), 1];
});

// How many points to plot between nodes
var tstep = 20;

var default_s = 1;

var dragged = null,
    selected = nodes[0];

var slider = d3.slider()
    .axis(true)
    .min(-1)
    .max(1)
    .step(0.01)
    .value(1)
    .on("slide", schange);

var line = d3.svg.line();

var svg = d3.select("#splines").append("svg")
    .attr("width", width)
    .attr("height", height)
    .attr("tabindex", 1);

svg.append("rect")
    .attr("width", width)
    .attr("height", height)
    .style("fill", "#EEEEEE")
    .on("mousedown", mousedown);

svg.append("path")
    .datum(nodes)
    .attr("class", "line")
    .call(redraw);

d3.select(window)
    .on("mousemove", mousemove)
    .on("mouseup", mouseup)
    .on("keydown", keydown);

d3.select("#interpolate")
    .on("change", change)
    .selectAll("option")
    .data([
        "linear",
        "step-before",
        "step-after",
        "basis",
        "basis-open",
        "basis-closed",
        "cardinal",
        "cardinal-open",
        "cardinal-closed",
        "monotone"
    ])
    .enter().append("option")
    .attr("value", function(d) { return d; })
    .text(function(d) { return d; });

svg.node().focus();

d3.select("#slider").call(slider);

var closed_xspline = false;

// Parameter space, "time", one "second" per node
var tvec = linspace(0, nodes.length-1, (nodes.length-1)*tstep + 1);

var xspline = d3.svg.line()
    .x(function(d,i){return xspline_at_t(d, 0);})
    .y(function(d,i){return xspline_at_t(d, 1);});

svg.append('path')
    .datum(tvec)
    .attr('class','xspline')
    .call(redraw);

function redraw() {
    svg.select("path.line").attr("d", line);
    if (!closed_xspline) {
        tvec = linspace(0, nodes.length-1, (nodes.length-1)*tstep + 1);
    } else {
        tvec = linspace(0, nodes.length, nodes.length*tstep + 1);
    }
    svg.select("path.xspline").datum(tvec).attr("d", xspline);

    var circle = svg.selectAll("circle")
        .data(nodes, function(d) { return d; });

    circle.enter().append("circle")
        .attr("r", 1e-6)
        .on("mousedown", function(d, i) {
            selected = dragged = d;
            slider.value(d[2]);
            redraw();
        })
        .transition()
        .duration(750)
        .ease("elastic")
        .attr("r", 6.5);

    circle
        .classed("selected", function(d) { return d === selected; })
        .attr("cx", function(d) { return d[0]; })
        .attr("cy", function(d) { return d[1]; });

    circle.exit().remove();

    if (d3.event) {
        d3.event.preventDefault();
        d3.event.stopPropagation();
    }
}

function change() {
    line.interpolate(this.value);
    redraw();
}

function schange(evt, value){
    selected[2] = value;
    svg.select("path.xspline").attr("d", xspline);
}

function mousedown() {
    var newcirc = d3.mouse(svg.node());
    newcirc.push(default_s);
    slider.value(default_s);
    nodes.push(selected = dragged = newcirc);
    redraw();
}

function mousemove() {
    if (!dragged) return;
    var m = d3.mouse(svg.node());
    dragged[0] = Math.max(0, Math.min(width, m[0]));
    dragged[1] = Math.max(0, Math.min(height, m[1]));
    redraw();
}

function mouseup() {
    if (!dragged) return;
    mousemove();
    dragged = null;
}

function keydown() {
    if (!selected) return;
    switch (d3.event.keyCode) {
    case 8: // backspace
    case 46: { // delete
        var i = nodes.indexOf(selected);
        nodes.splice(i, 1);
        selected = nodes.length ? nodes[i > 0 ? i - 1 : 0] : null;
        redraw();
        break;
    }
    }
}

// Xspline stuff starts here

function linspace(a, b, n) {
    let stepsize = (b - a)/(n - 1);
    return [...Array(n).keys()].map(x => a + x*stepsize);
}

// Basic polynomials
function f(p, u) {
    return u*u*u*(10-p + u*(2*p-15 + u*(6-p)));
}

function g(q, u) {
    return u*(q + u*(2*q + u*((8 - 12*q) + u*(14*q - 11 + u*(4 - 5*q)))));
}

function h(q, u) {
    return q*u*(1 + u*(2  - u*u*(2 + u)));
}

// Non-normalised blending function at node k
function fk(k, t) {
    var n = nodes.length;

    if (! closed_xspline) {
        // No blending functions outside the range
        if(k >= n || k < 0) {
            return 0;
        }
    }


    // Determine if we're doing the left or right side
    var s, tboundary;
    if (t < k)
    {
        var leftk = (n + k - 1)%n;
        s = nodes[leftk][2];
        tboundary = k - 1 - s;
    }
    else {
        var rightk = (n + k + 1)%n;
        s = nodes[rightk][2];
        tboundary = k + 1 + s;
    }

    // Approximation or sharp interpolation at this node
    if (s >= 0) {
        // Check if it's outside the support of the function
        if ((t > k && t > tboundary) || (t < k && t < tboundary)) {
            return 0;
        }

        var p = 2*(1+s)*(1+s);
        return f(p, (t-tboundary)/(k-tboundary));
    }

    // Else, s < 0, and we're smoothly interpolating at this node
    var q = -s/2;

    //Outside of support
    if (t < k-2) {
        return 0;
    }
    //left negative
    else if (t < k-1) {
        return h(q, t - k + 1);
    }
    //left positive
    else if (t < k) {
        return g(q, t - k + 1);
    }
    //right positive
    else if (t < k+1) {
        return g(q, k + 1 -t);
    }
    // right negative
    else if (t < k+ 2){
        return h(q, k + 1 - t);
    }

    // Outside of support
    return 0;
}


function xspline_at_t(t, XorY) {
    // Determine which blending functions have influence at this point.
    var k1, k2, k3, k4;
    var tfloor = Math.floor(t);
    var tceil = Math.ceil(t);
    if ( t == tfloor) {
        // When exactly on a node, only immediate neighbours have
        // influence
        k1 = t - 1;
        k2 = t;
        k3 = t + 1;
        k4 = null;
    }
    else {
        k1 = tfloor - 1;
        k2 = tfloor;
        k3 = tceil;
        k4 = tceil + 1;
    }

    var n = nodes.length;

    // When indexing, it has to be mod n
    var k1_idx = (n + k1) % n,
        k2_idx = (n + k2) % n,
        k3_idx = (n + k3) % n,
        k4_idx = k4 != null ? (n + k4) % n : null;

    var f1t = fk(k1, t),
        f2t = fk(k2, t),
        f3t = fk(k3, t),
        f4t = k4 != null ? fk(k4, t) : 0;

    // normalisation factor
    var denom = f1t + f2t + f3t + f4t;

    var p1 = nodes[k1_idx],
        p2 = nodes[k2_idx],
        p3 = nodes[k3_idx],
        p4 = k4 != null ? nodes[k4_idx] : [0, 0];

    var at_t = (f1t*p1[XorY] + f2t*p2[XorY] + f3t*p3[XorY] + f4t*p4[XorY])/denom;

    return at_t;
}

function set_all_s() {
    for(var i = 0; i < nodes.length; i++) {
        nodes[i][2] = default_s;
    }
}

function approximate_all() {
    default_s = 1;
    set_all_s();
    redraw();
}

function sharply_interpolate_all() {
    default_s = 0;
    set_all_s();
    redraw();
}

function interpolate_all() {
    default_s = -1;
    set_all_s();
    redraw();
}

function toggle_closed() {
    closed_xspline = !closed_xspline;
    redraw();
}
